/***************************************************************************
                          net.c  -  description
                             -------------------
    begin                : t z 18 2003
    copyright            : (C) 2003 by Petr Cermak
    email                : cermak01@seznam.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "net.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

 

/**
 * Waits for single client to connect
 *
 * @param port On which port to wait
 * @return Descriptor of created socked, or -1 if unsuccesfull
 */
int serverAccept(int port)
{
	int listenSock, newSock;
	socklen_t sz;
	struct protoent* pp;
	struct sockaddr_in sa, ca;

	errno = 0;
	if (port > 65535) {
		return -1;
	}
	 
	pp = getprotobyname("tcp");
	listenSock = socket(AF_INET, SOCK_STREAM, pp->p_proto);

	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sa.sin_addr.s_addr = INADDR_ANY;

	if (bind(listenSock, (struct sockaddr*) &sa, sizeof(sa)) == -1) {
		return -1;
	}

	if (listen(listenSock, 1) == -1) {
		close(listenSock);
		return -1;
	}

	sz = sizeof(ca);
	newSock = accept(listenSock, (struct sockaddr*) &ca, &sz);
	if (close(listenSock) == -1) {
		close(newSock);
		return -1;
	}
	
	return newSock; 
}

/**
 * Connects to server
 *
 * @param host Address / hostname of server
 * @param port Port on which server listens
 * @return Connected socket, or -1 if unsuccesfull
 */
int clientConnect(char* host, int port)
{
	int newSock;
	struct protoent* pp;
	struct hostent* ha;
	struct sockaddr_in sa;

	h_errno = 0;
	errno = 0;
	if (port > 65535) {
		return -1;
	}
	
	pp = getprotobyname("tcp");
	h_errno = 0;
	ha = gethostbyname(host);
	if (ha == NULL) {
		return -1;
	}

	newSock = socket(AF_INET, SOCK_STREAM, pp->p_proto);

	sa.sin_family  = AF_INET;
	sa.sin_port = htons(port);
	memcpy(&sa.sin_addr.s_addr, ha->h_addr_list[0], ha->h_length);
	
	if (connect(newSock, (struct sockaddr*) &sa, sizeof(sa)) == -1) {
		return -1;
	}

	return newSock;
}

/**
 * Reads a game from stream
 *
 * @param fd File destriptor where to store game into
 * @return Pointer to game read, or NULL if unsuccesfull
 */
Game* readGame(int fd, Boolean useRecv)
{
	int j, sz, sz1;
	Game* netGame = (Game*) malloc(sizeof(Game));
	Game* g = (Game*) malloc(sizeof(Game));

	if (!g || !netGame) {
		return 0;
	}

	errno = 0;
	if (useRecv) {
		sz = recv(fd, netGame, sizeof(Game), MSG_WAITALL);
	} else {
		sz = read(fd, netGame, sizeof(Game));
	}

	if (sz < sizeof(Game)) {
		free(g);
		free(netGame);
		return 0;
	}

	g->rows = ntohl(netGame->rows);
	g->cols = ntohl(netGame->cols);
	g->empty = ntohl(netGame->empty);
	g->gs = netGame->gs;
	g->whoAmI = netGame->whoAmI;
	g->redScore = ntohl(netGame->redScore);
	g->blueScore = ntohl(netGame->blueScore);
	g->pisLen = ntohl(netGame->pisLen);

	free(netGame);

	g->fields = (FieldState**) malloc(sizeof(FieldState*) * g->rows);
	for (j = 0; j < g->rows; j++) {
		sz1 = sizeof(FieldState) * g->cols;
		errno = 0;
		if (!(g->fields[j] = (FieldState*) malloc(sz1))) {
			return 0;
		};

		// no need to translate to net format. FieldState is char
		if (useRecv) {
			sz = recv(fd, g->fields[j], sz1, MSG_WAITALL);
		} else {
			sz = read(fd, g->fields[j], sz1);
		}

		if (sz < sz1) {
			return 0;
		}
	}

	return g;
}

/**
 * Stores a game into stream
 *
 * @param fd File destriptor where to store game into
 * @param g Game to store
 * @return TRUE - ok, FALSE - error
 */
Boolean storeGame(int fd, Game* g)
{
	int j, sz;
	Game* netGame = (Game*) malloc(sizeof(Game));

	if (!netGame) {
		return FALSE;
	}

	netGame->rows = htonl(g->rows);
	netGame->cols = htonl(g->cols);
	netGame->empty = htonl(g->empty);
	netGame->gs = g->gs;
	netGame->whoAmI = g->whoAmI;
	netGame->redScore = htonl(g->redScore);
	netGame->blueScore = htonl(g->blueScore);
	netGame->pisLen = htonl(g->pisLen);


	errno = 0;
	if (write(fd, netGame, sizeof(Game)) < sizeof(Game)) {
		free(netGame);
		return FALSE;
	}

	free(netGame);

	for (j = 0; j < g->rows; j++) {
		sz = sizeof(FieldState) * g->cols;
		errno = 0;
		// no need to translate to net format. FieldState is char
		if (write(fd, g->fields[j], sz) < sz) {
			return FALSE;
		}
	}

	return TRUE;
}

/**
 * Saves game to a file.
 * Given file will be overwritten
 *
 * @param fileName Name of file where to store game into.
 * @param g Game to store
 */
Boolean saveGameToFile(char* fileName, Game* g)
{
	int fd;

	if ((fd = open(fileName, O_CREAT | O_WRONLY | O_TRUNC, 0666)) == -1) {
		return FALSE;
	}

	if (!storeGame(fd, g)) {
		close(fd);
		return FALSE;
	}

	if (close(fd) == -1) {
		return FALSE;
	}

	return TRUE;
}

/**
 * Loads game from a file.
 *
 * @param fileName Filename of saved game
 * @return Pointer to a new game loaded, or NULL if unsuccesfull
 */
Game* loadGameFromFile(char* fileName)
{
	int fd;
	Game* retGame;

	if ((fd = open(fileName, O_RDONLY)) == -1) {
		return NULL;
	}

	retGame = readGame(fd, FALSE);

	if (close(fd) == -1) {
		deleteGame(retGame);
		return NULL;
	}

	return retGame;
}

/**
 * Sends message to socket fd
 *
 * @param fd Socket where to send message into
 * @param msg Message to send
 * @return TRUE - successfull, FALSE - otherwise
 */
Boolean recvNetMessage(int fd, NetMessage* msg)
{
	int sz;

	errno = 0;
	sz = recv(fd, msg, sizeof(NetMessage), MSG_WAITALL);

	if (sz < sizeof(NetMessage)) {
		return FALSE;
	}

	msg->type = ntohl(msg->type);
	msg->x = ntohl(msg->x);
	msg->y = ntohl(msg->y);

	return TRUE;	
}

/**
 * Receives message from socket fd
 *
 * @param fd Socket to receive message from
 * @param msg Message to be received
 * @return TRUE - successfull, FALSE - otherwise
 */
Boolean sendNetMessage(int fd, NetMessage* msg)
{
	int sz;

	msg->type = htonl(msg->type);
	msg->x = htonl(msg->x);
	msg->y = htonl(msg->y);
	
	sz = send(fd, msg, sizeof(NetMessage), 0);
	if (sz < sizeof(NetMessage)) {
		return FALSE;
	}

	return TRUE;
}

/**
 * Sends text message
 *
 * @param fd Socket where to send message into
 * @param msg Text message
 * @return TRUE - successfull, FALSE - otherwise
 */
Boolean sendCommandMsg(int fd, char* msg)
{
	NetMessage netMsg;

	memset(&netMsg, 0, sizeof(NetMessage));
	netMsg.type = NM_MSG;
	strncpy(netMsg.str, msg, 256);
	netMsg.str[255] = '\0';
	return sendNetMessage(fd, &netMsg);
}

/**
 * Sends play at message
 *
 * @param fd Socket where to send message into
 * @param x Where to play
 * @param y Where to play
 * @return TRUE - successfull, FALSE - otherwise
 */
Boolean sendCommandPlay(int fd, int x, int y)
{
	NetMessage netMsg;

	memset(&netMsg, 0, sizeof(NetMessage));
	netMsg.type = NM_PLAYAT;
	netMsg.x = x;
	netMsg.y = y;
	return sendNetMessage(fd, &netMsg);
}

/**
 * Sends game message
 *
 * @param fd Socket where to send message into
 * @param type Game message to send
 * @return TRUE - successfull, FALSE - otherwise
 */
Boolean sendCommand(int fd, NetMsgType type)
{
	NetMessage netMsg;

	memset(&netMsg, 0, sizeof(NetMessage));
	netMsg.type = type;
	return sendNetMessage(fd, &netMsg);
}
